@@ -45,6 +45,10 @@ private slots:
     void muxerWriteFrames();
     void muxerWriteSubtitles();
     void muxerEnqueue();
+    void muxerEnqueueFramesFromMultiSources();
+    void muxerEnqueueFramesFromDev();
+    void muxerWritePacketsFromMultiSources();
+    void muxerWritePacketsFromDev();
 };
 
 void tst_QAVDemuxer::construction()
@@ -56,16 +60,17 @@ void tst_QAVDemuxer::construction()
     QCOMPARE(d.duration(), 0);
     QCOMPARE(d.seekable(), false);
     QCOMPARE(d.eof(), false);
-    QVERIFY(!d.read());
+    QAVPacket p;
+    QVERIFY(!p);
+    QVERIFY(d.read(p) < 0);
     QVERIFY(d.seek(0) < 0);
 
-    QAVPacket p;
     QVERIFY(!p);
     QCOMPARE(p.duration(), 0);
     QCOMPARE(p.packet()->size, 0);
     QVERIFY(p.packet()->stream_index < 0);
     QList<QAVFrame> fs;
-    d.decode(p, fs);
+    QAVDemuxer::decode(p, fs);
     QVERIFY(fs.isEmpty());
 
     QAVFrame f;
@@ -87,7 +92,8 @@ void tst_QAVDemuxer::loadIncorrect()
     QVERIFY(d.currentVideoStreams().isEmpty());
     QVERIFY(d.currentAudioStreams().isEmpty());
     QVERIFY(d.currentSubtitleStreams().isEmpty());
-    QVERIFY(!d.read());
+    QAVPacket p;
+    QVERIFY(d.read(p) < 0);
     QVERIFY(d.seek(0) < 0);
 }
 
@@ -108,7 +114,7 @@ void tst_QAVDemuxer::loadAudio()
     QAVPacket p;
     QAVPacket p2;
     QAVFrame f2;
-    while ((p = d.read())) {
+    while (d.read(p) >= 0) {
         QVERIFY(p);
         QVERIFY(p.packet());
         QVERIFY(p.duration() > 0);
@@ -129,7 +135,7 @@ void tst_QAVDemuxer::loadAudio()
         QCOMPARE(p2.packet()->stream_index, p.packet()->stream_index);
 
         QList<QAVFrame> fs;
-        d.decode(p, fs);
+        QAVDemuxer::decode(p, fs);
         QCOMPARE(fs.size(), 1);
         auto f = fs[0];
         QVERIFY(f);
@@ -188,7 +194,8 @@ void tst_QAVDemuxer::loadVideo()
     QCOMPARE(d.seekable(), true);
     QCOMPARE(d.eof(), false);
 
-    auto p = d.read();
+    QAVPacket p;
+    QVERIFY(d.read(p) >= 0);
     QVERIFY(p);
     QVERIFY(p.packet());
     QVERIFY(p.duration() > 0);
@@ -196,7 +203,7 @@ void tst_QAVDemuxer::loadVideo()
     QVERIFY(p.packet()->stream_index >= 0);
 
     QList<QAVFrame> fs;
-    d.decode(p, fs);
+    QAVDemuxer::decode(p, fs);
     QCOMPARE(fs.size(), 1);
     auto f = fs[0];
     QVERIFY(f);
@@ -205,11 +212,11 @@ void tst_QAVDemuxer::loadVideo()
     QCOMPARE(d.eof(), false);
 
     QVERIFY(d.seek(0) >= 0);
-    while ((p = d.read())) {
+    while (d.read(p) >= 0) {
         QCOMPARE(d.eof(), false);
         if (p.packet()->stream_index == d.currentVideoStreams().first().index()) {
             QList<QAVFrame> fs1;
-            d.decode(p, fs1);
+            QAVDemuxer::decode(p, fs1);
             QCOMPARE(fs1.size(), 1);
             f = fs1[0];
             QVERIFY(f);
@@ -247,7 +254,8 @@ void tst_QAVDemuxer::fileIO()
     QCOMPARE(d.seekable(), true);
     QCOMPARE(d.eof(), false);
 
-    auto p = d.read();
+    QAVPacket p;
+    QVERIFY(d.read(p) >= 0);
     QVERIFY(p);
     QVERIFY(p.packet());
     QVERIFY(p.duration() > 0);
@@ -255,7 +263,7 @@ void tst_QAVDemuxer::fileIO()
     QVERIFY(p.packet()->stream_index >= 0);
 
     QList<QAVFrame> fs;
-    d.decode(p, fs);
+    QAVDemuxer::decode(p, fs);
     QCOMPARE(fs.size(), 1);
     auto f = fs[0];
     QVERIFY(f);
@@ -264,11 +272,11 @@ void tst_QAVDemuxer::fileIO()
     QCOMPARE(d.eof(), false);
 
     QVERIFY(d.seek(0) >= 0);
-    while ((p = d.read())) {
+    while (d.read(p) >= 0) {
         QCOMPARE(d.eof(), false);
         if (p.packet()->stream_index == d.currentVideoStreams().first().index()) {
             QList<QAVFrame> fs1;
-            d.decode(p, fs1);
+            QAVDemuxer::decode(p, fs1);
             QCOMPARE(fs1.size(), 1);
             f = fs1[0];
             QVERIFY(f);
@@ -308,7 +316,7 @@ void tst_QAVDemuxer::qrcIO()
     QAVPacket p;
     QAVPacket p2;
     QAVFrame f2;
-    while ((p = d.read())) {
+    while (d.read(p) >= 0) {
         QVERIFY(p);
         QVERIFY(p.packet());
         QVERIFY(p.duration() > 0);
@@ -329,7 +337,7 @@ void tst_QAVDemuxer::qrcIO()
         QCOMPARE(p2.packet()->stream_index, p.packet()->stream_index);
 
         QList<QAVFrame> fs;
-        d.decode(p, fs);
+        QAVDemuxer::decode(p, fs);
         QCOMPARE(fs.size(), 1);
         auto f = fs[0];
         QVERIFY(f);
@@ -449,7 +457,7 @@ void tst_QAVDemuxer::muxerWritePackets()
     QVERIFY(m.load(d.availableStreams(), "colors.mkv") >= 0);
 
     QAVPacket p;
-    while ((p = d.read())) {
+    while (d.read(p) >= 0) {
         QVERIFY(m.write(p) >= 0);
     }
     QVERIFY(m.flush() >= 0);
@@ -468,9 +476,9 @@ void tst_QAVDemuxer::muxerWriteFrames()
     QVERIFY(m.load(d.availableStreams(), "colors.mkv") >= 0);
 
     QAVPacket p;
-    while ((p = d.read())) {
+    while (d.read(p) >= 0) {
         QList<QAVFrame> fs;
-        d.decode(p, fs);
+        QAVDemuxer::decode(p, fs);
         if (fs.size()) {
             QVERIFY(fs.size() == 1);
             auto &f = fs[0];
@@ -493,12 +501,12 @@ void tst_QAVDemuxer::muxerWriteSubtitles()
     QVERIFY(m.load(d.availableStreams(), "colors.mkv") >= 0);
 
     QAVPacket p;
-    while ((p = d.read())) {
+    while (d.read(p) >= 0) {
         switch (p.stream().codec()->avctx()->codec_type) {
         case AVMEDIA_TYPE_VIDEO:
         case AVMEDIA_TYPE_AUDIO: {
             QList<QAVFrame> fs;
-            d.decode(p, fs);
+            QAVDemuxer::decode(p, fs);
             if (fs.size()) {
                 QVERIFY(fs.size() == 1);
                 auto &f = fs[0];
@@ -508,7 +516,7 @@ void tst_QAVDemuxer::muxerWriteSubtitles()
         }
         case AVMEDIA_TYPE_SUBTITLE: {
             QList<QAVSubtitleFrame> fs;
-            d.decode(p, fs);
+            QAVDemuxer::decode(p, fs);
             if (fs.size()) {
                 QVERIFY(fs.size() == 1);
                 auto &f = fs[0];
@@ -537,9 +545,9 @@ void tst_QAVDemuxer::muxerEnqueue()
     QVERIFY(m.load(d.availableStreams(), "colors.mkv") >= 0);
 
     QAVPacket p;
-    while ((p = d.read())) {
+    while (d.read(p) >= 0) {
         QList<QAVFrame> fs;
-        d.decode(p, fs);
+        QAVDemuxer::decode(p, fs);
         if (fs.size()) {
             QVERIFY(fs.size() == 1);
             auto &f = fs[0];
@@ -553,5 +561,157 @@ void tst_QAVDemuxer::muxerEnqueue()
     QVERIFY(d.load("colors.mkv") >= 0);
 }
 
+void tst_QAVDemuxer::muxerEnqueueFramesFromMultiSources()
+{
+    QFileInfo file1(testData("colors.mp4"));
+    QFileInfo file2(testData("small.mp4"));
+    QAVDemuxer d1;
+    QAVDemuxer d2;
+    QAVMuxerFrames m;
+
+    QVERIFY(d1.load(file1.absoluteFilePath()) >= 0);
+    QVERIFY(d2.load(file2.absoluteFilePath()) >= 0);
+    auto streams = d1.availableStreams() + d2.availableStreams();
+    QVERIFY(m.load(streams, "colors.mkv") >= 0);
+
+    auto decode = [&](auto &p) {
+        if (!p.stream())
+            return;
+        QList<QAVFrame> fs;
+        QAVDemuxer::decode(p, fs);
+        if (fs.size()) {
+            QVERIFY(fs.size() == 1);
+            auto &f = fs[0];
+            m.enqueue(f);
+        }
+    };
+
+    QAVPacket p1;
+    QAVPacket p2;
+    while (d1.read(p1) >= 0 || d2.read(p2) >= 0) {
+        decode(p1);
+        decode(p2);
+    }
+    QVERIFY(m.flush() >= 0);
+    m.unload();
+    d1.unload();
+    d2.unload();
+    QVERIFY(d1.load("colors.mkv") >= 0);
+    QVERIFY(d1.availableStreams().size() == 4);
+    QVERIFY(d1.availableVideoStreams().size() == 2);
+    QVERIFY(d1.availableAudioStreams().size() == 2);
+}
+
+void tst_QAVDemuxer::muxerEnqueueFramesFromDev()
+{
+    QAVDemuxer d1;
+    QAVDemuxer d2;
+    QAVMuxerFrames m;
+    auto fmts = QAVDemuxer::supportedFormats();
+    QVERIFY(!fmts.isEmpty());
+    QVERIFY(!QAVDemuxer::supportedProtocols().isEmpty());
+    if (!fmts.contains(QLatin1String("v4l2")))
+        return;
+    QFileInfo file1(QLatin1String("/dev/video0"));
+    if (!file1.exists())
+        return;
+
+    d1.setInputFormat(QLatin1String("v4l2"));
+    QVERIFY(d1.load(QLatin1String("/dev/video0")) >= 0);
+    QFileInfo file2(testData("small.mp4"));
+    QVERIFY(d2.load(file2.absoluteFilePath()) >= 0);
+    auto streams = d1.availableStreams() + d2.availableStreams();
+    QVERIFY(m.load(streams, "output.mkv") >= 0);
+
+    auto decode = [&](auto &p) {
+        if (!p.stream())
+            return;
+        QList<QAVFrame> fs;
+        QAVDemuxer::decode(p, fs);
+        if (fs.size()) {
+            QVERIFY(fs.size() == 1);
+            auto &f = fs[0];
+            m.enqueue(f);
+        }
+    };
+
+    QAVPacket p1;
+    QAVPacket p2;
+    while (d1.read(p1) >= 0 && d2.read(p2) >= 0) {
+        decode(p1);
+        decode(p2);
+    }
+
+    QVERIFY(m.flush() >= 0);
+    m.unload();
+    p1 = {};
+    p2 = {};
+    d1.unload();
+    d2.unload();
+    QAVDemuxer d;
+    QVERIFY(d.load("output.mkv") >= 0);
+    QVERIFY(d.availableStreams().size() == 3);
+    QVERIFY(d.availableVideoStreams().size() == 2);
+    QVERIFY(d.availableAudioStreams().size() == 1);
+}
+
+void tst_QAVDemuxer::muxerWritePacketsFromMultiSources()
+{
+    QFileInfo file1(testData("colors.mp4"));
+    QFileInfo file2(testData("small.mp4"));
+    QAVDemuxer d1;
+    QAVDemuxer d2;
+    QAVMuxerPackets m;
+
+    QVERIFY(d1.load(file1.absoluteFilePath()) >= 0);
+    QVERIFY(d2.load(file2.absoluteFilePath()) >= 0);
+    auto streams = d1.availableStreams() + d2.availableStreams();
+    QVERIFY(m.load(streams, "colors.mkv") >= 0);
+
+    QAVPacket p1;
+    QAVPacket p2;
+    while (d1.read(p1) >= 0 || d2.read(p2) >= 0) {
+        m.write(p1);
+        m.write(p2);
+    }
+    QVERIFY(m.flush() >= 0);
+    m.unload();
+    d1.unload();
+    d2.unload();
+    QVERIFY(d1.load("colors.mkv") >= 0);
+    QVERIFY(d1.availableStreams().size() == 4);
+    QVERIFY(d1.availableVideoStreams().size() == 2);
+    QVERIFY(d1.availableAudioStreams().size() == 2);
+}
+
+void tst_QAVDemuxer::muxerWritePacketsFromDev()
+{
+    QAVDemuxer d1;
+    QAVMuxerPackets m;
+    auto fmts = QAVDemuxer::supportedFormats();
+    QVERIFY(!fmts.isEmpty());
+    QVERIFY(!QAVDemuxer::supportedProtocols().isEmpty());
+    if (!fmts.contains(QLatin1String("v4l2")))
+        return;
+    QFileInfo file1(QLatin1String("/dev/video0"));
+    if (!file1.exists())
+        return;
+
+    d1.setInputFormat(QLatin1String("v4l2"));
+    d1.setInputVideoCodec("h264");
+    QVERIFY(d1.load(QLatin1String("/dev/video0")) >= 0);
+    auto streams = d1.availableStreams();
+    QVERIFY(m.load(streams, "output.yuv") >= 0);
+
+    QAVPacket p1;
+    int i = 30;
+    while (d1.read(p1) >= 0 && i-- > 0) {
+        m.write(p1);
+    }
+
+    QVERIFY(m.flush() >= 0);
+    // ffmpeg -f rawvideo -pix_fmt yuyv422 -s:v 640x480 -r 25 -i output.yuv -c:v libx264 output.mp4
+}
+
 QTEST_MAIN(tst_QAVDemuxer)
 #include "tst_qavdemuxer.moc"
